package org.insa.graphs.algorithm.shortestpath;

import java.util.ArrayList;
import java.util.Collections;

import org.insa.graphs.algorithm.AbstractSolution.Status;
import org.insa.graphs.algorithm.utils.BinaryHeap;
import org.insa.graphs.model.Arc;
import org.insa.graphs.model.Graph;
import org.insa.graphs.model.Node;
import org.insa.graphs.model.Path;

public class DijkstraAlgorithm extends ShortestPathAlgorithm {

    public DijkstraAlgorithm(ShortestPathData data) {
        super(data);
    }

    @Override
    protected ShortestPathSolution doRun() {
        final ShortestPathData data = getInputData();
        ShortestPathSolution solution = null;
        // TODO:

        // List of sommets, but with type Label
        ArrayList<Label> labels = new ArrayList<Label>(); 
        // Heap of sommets
        BinaryHeap<Label> tas = new BinaryHeap<Label>();

        Graph graph = data.getGraph();
        final int nbNodes = graph.size();

        // Notify observers about the first event (origin processed).
        notifyOriginProcessed(data.getOrigin());

        // Init Dijkstra
        for (Node node : graph.getNodes()) {
            labels.add(new Label(node));
        }
        
        Label s = new Label(data.getOrigin());
        
        // Add origin in the heap
        //Label s = origin;//labels.get(0);
        s.setPathCost(0);
        tas.insert(s);

        // On considère le sommet x à chaque itération
        Label x;

        while (!tas.isEmpty()) {
            x = tas.deleteMin();
            x.mark();

            // A marked node is considered as reached
            notifyNodeReached(x.getNode());

            // We create a list of node successors of x, instead of a list of Arcs.
            float arc_cost = 0;
            ArrayList<Node> successors = new ArrayList<Node>();
            for (Arc k : x.getNode().getSuccessors()) {
                successors.add(k.getDestination());
            }

            // We loop only on the successors of the node with this for and if.
            for (Label y : labels) {
                if (successors.contains(y.getNode())) {
                    if (!y.isMarked()) {
                        // This loop serves to get the lentgh of the arc as
                        // we know its origin and destination
                        for (Arc arc : x.getNode().getSuccessors()) {
                            if (y.getNode() == arc.getDestination()) {
                                arc_cost = arc.getLength();
                            }
                        }
                        float possible_path_cost = x.getCost() + arc_cost;
                        if (y.getCost() > possible_path_cost) {
                            // Mise à jour du label
                            y.setPathCost(possible_path_cost);
                            y.setParentNode(x.getNode());
                            // Si y est déjà dans le tas, on va l'update, pas l'insérer
                            // On a pas de fonction update donc on fait un remove puis insert
                            if (tas.array.contains(y)){
                                tas.remove(y);
                                tas.insert(y);
                            }
                            // il n'y est pas déjà on l'insère
                            else {
                                tas.insert(y);
                            }
                        }
                    }
            }
            }
        }

        // Create the path ...
        ArrayList<Arc> arcs_path = new ArrayList<>();
    
        // We will find the path using the parent nodes, from the destination to the origin
        Node current_node = data.getDestination();
        Label current_label = new Label(current_node);
        System.out.println(current_node.getId());
    
        while (current_node != null && current_node != data.getOrigin()) {
            // Find the label matching the parent node
            for (Label label : labels) {
                if (label.getNode() == current_node){
                    current_label = label;
                    current_node = current_label.getParentNode();
                }
            }
            // Knowing the parent node, get the arc between the parent and
            // current node and add it to the path
            if (current_node != null){
                for (Arc arc : current_node.getSuccessors()) {
                    if (arc.getDestination().getId() == current_label.getNode().getId() ) {
                        arcs_path.add(arc);
                        break;
                    }
                }
            }
        }
            notifyDestinationReached(data.getDestination());

        

        // Reverse the path...
        Collections.reverse(arcs_path);

        // Create the final solution.
        solution = new ShortestPathSolution(data, Status.OPTIMAL, new Path(graph, arcs_path));

        /*System.out.println("watch here" + "\n");
        for (Arc a : arcs_path){
            System.out.println(a.getOrigin().getId() + " --- " + a.getDestination().getId() + "\n");
        }*/
        return solution;
    }

}
